#'@rdname CST_ProxiesAttractor
#'@title Computing two dinamical proxies of the attractor in s2dv_cube.
#'
#'@author Carmen Alvarez-Castro, \email{carmen.alvarez-castro@cmcc.it}
#'@author Maria M. Chaves-Montero, \email{mdm.chaves-montero@cmcc.it}
#'@author Veronica Torralba, \email{veronica.torralba@cmcc.it}
#'@author Davide Faranda, \email{davide.faranda@lsce.ipsl.fr}
#'
#'@description This function computes two dinamical proxies of the attractor: 
#'The local dimension (d) and the inverse of the persistence (theta) for an
#''s2dv_cube' object. These two parameters will be used as a condition for the 
#'computation of dynamical scores to measure predictability and to compute bias 
#'correction conditioned by the dynamics with the function DynBiasCorrection.  
#'Funtion based on the matlab code (davide.faranda@lsce.ipsl.fr) used in the 
#'articles Faranda et al. (2017) and Faranda et al. (2019).  
#'@references Faranda, D., Alvarez-Castro, M.C., Messori, G., Rodriguez, D., and 
#'Yiou, P. (2019). The hammam effect or how a warm ocean enhances large scale 
#'atmospheric predictability. Nature Communications, 10(1), 1316. 
#'DOI = https://doi.org/10.1038/s41467-019-09305-8".
#'@references Faranda, D., Gabriele Messori and Pascal Yiou. (2017).
#'Dynamical proxies of North Atlantic predictability and extremes. 
#'Scientific Reports, 7-41278, 2017.
#'
#'@param data An 's2dv_cube' object with the data to create the attractor. 
#'  The data element should be an array with named spatial and temporal 
#'  dimensions.
#'@param quanti A numeric value lower than 1 indicating the quantile to perform 
#'  the computation of local dimension (d) and the the inverse of the  
#'  persistence (theta).
#'@param spatial_dims A vector of character strings indicating the name of the 
#'  spatial dimensions. The default is c('lat', 'lon'). They can be NULL if 
#'  the data is not spatial. If they are more than 2, they will be merged.
#'@param time_dims A vector of character strings indicating the name of the 
#'  temporal dimensions. The default is c('ftime', 'sdate'). If they are more 
#'  than 2, they will be merged.  
#'@param na.rm A logical value indicating whether to remove the NAs when 
#'  calculating the logaritmic distance. By default it is set to TRUE.
#'@param ncores The number of cores to use in parallel computation. It is set 
#'  to NULL by default, which means that the number of cores is 1.
#' 
#'@return An 's2dv_cube' object with the element data containing a list with 
#'the local dimension (dim) and the inverse of the persistence (theta) of the 
#'attractor.
#'
#'@examples
#'# Example 1: Computing the attractor using simple s2dv data
#'attractor <- CST_ProxiesAttractor(data = lonlat_temp$obs, quanti = 0.6)
#'
#'@import multiApply
#'@export
CST_ProxiesAttractor <- function(data, quanti, spatial_dims = c('lat', 'lon'),
                                 time_dims = c('ftime', 'sdate'), 
                                 na.rm = TRUE, ncores = NULL) {

  if (!inherits(data, 's2dv_cube')) {
    stop("Parameter 'data' must be of the class 's2dv_cube'.")
  }
  data$data <- ProxiesAttractor(data = data$data, quanti = quanti, 
                                spatial_dims = spatial_dims, 
                                time_dims = time_dims, 
                                na.rm = na.rm, ncores = ncores)
  return(data)
}
#'@rdname ProxiesAttractor
#'
#'@title Computing two dinamical proxies of the attractor
#'@author Carmen Alvarez-Castro, \email{carmen.alvarez-castro@cmcc.it}
#'@author Maria M. Chaves-Montero, \email{mdm.chaves-montero@cmcc.it}
#'@author Veronica Torralba, \email{veronica.torralba@cmcc.it}
#'@author Davide Faranda, \email{davide.faranda@lsce.ipsl.fr}
#'
#'@description This function computes two dinamical proxies of the attractor: 
#'The local dimension (d) and the inverse of the persistence (theta) for an
#''s2dv_cube' object. These two parameters will be used as a condition for the 
#'computation of dynamical scores to measure predictability and to compute bias 
#'correction conditioned by the dynamics with the function DynBiasCorrection.  
#'Funtion based on the matlab code (davide.faranda@lsce.ipsl.fr) used in the 
#'articles Faranda et al. (2017) and Faranda et al. (2019).  
#'@references Faranda, D., Alvarez-Castro, M.C., Messori, G., Rodriguez, D., 
#'and Yiou, P. (2019). The hammam effect or how a warm ocean enhances large 
#'scale atmospheric predictability. Nature Communications, 10(1), 1316. 
#'DOI = https://doi.org/10.1038/s41467-019-09305-8 "
#'@references Faranda, D., Gabriele Messori and Pascal Yiou. (2017).
#'Dynamical proxies of North Atlantic predictability and extremes. 
#'Scientific Reports, 7-41278, 2017. 
#'
#'@param data A multidimensional array with named dimensions to create the 
#'  attractor. It requires a temporal dimension and spatial dimensions: 
#'  longitudinal and latitudinal, or grid.
#'@param quanti A numeric value lower than 1 indicating the quantile to perform 
#'  the computation of local dimension and theta.
#'@param A vector of character strings indicating the name of the 
#'  spatial dimensions. The default is c('lat', 'lon'). They can be NULL if 
#'  the data is not spatial. If they are more than 2, they will be merged.
#'@param time_dims A vector of character strings indicating the name of the 
#'  temporal dimensions. The default is c('ftime', 'sdate'). If they are more 
#'  than 2, they will be merged. 
#'@param na.rm A logical value indicating whether to remove the NAs when 
#'  calculating the logaritmic distance. By default it is set to TRUE.
#'@param ncores The number of cores to use in parallel computation. It is set 
#'  to NULL by default, which means that the number of cores is 1.
#'
#'@return A list with two arrays with the local dimension (dim) and the inverse 
#'of the persistence (theta) of the attractor.
#'
#'@examples
#'# Example 1: Computing the attractor using simple data
#'# Creating an example of matrix data(time,grids):
#'mat <- array(rnorm(36 * 40), c(time = 36, grid = 40)) 
#'qm <- 0.90 # imposing a threshold
#'Attractor <- ProxiesAttractor(data = mat, quanti = qm)
#'# to plot the result
#'time = c(1:length(Attractor$theta))
#'layout(matrix(c(1, 3, 2, 3), 2, 2))
#'plot(time, Attractor$dim, xlab = 'time', ylab = 'd',
#'     main = 'local dimension', type = 'l')
#'plot(time, Attractor$theta, xlab = 'time', ylab = 'theta', main = 'theta') 
#'plot(Attractor$dim, Attractor$theta, col = 'blue', 
#'     main = "Proxies of the Attractor",
#'     xlab = "local dimension", ylab = "theta", lwd = 8, 'p')
#'
#'@import multiApply
#'@export
ProxiesAttractor <- function(data, quanti, spatial_dims = c('lon', 'lat'), 
                             time_dims = c('ftime', 'sdate'), 
                             na.rm = TRUE, ncores = NULL) {
  # Initial checks
  ## data
  if (is.null(data)) {
    stop("Parameter 'data' cannot be NULL.")
  }
  dimnames <- names(dim(data))
  # quanti
  if (is.null(quanti)) {
    stop("Parameter 'quanti' cannot be NULL.")
  }
  # spatial_dims
  input_spatial_dims <- spatial_dims
  if (is.null(spatial_dims)) {
    data <- InsertDim(data, posdim = 1, lendim = 1, name = 'grid')
    spatial_dims <- 'grid'
  } else {
    if (!all(spatial_dims %in% names(dim(data)))) {
      stop("Parameter 'spatial_dims' is not found in 'data' dimension names. ", 
           "Please check the dimension names of 'data' or set it to NULL.")
    }
  }
  if (length(spatial_dims) == 2) {
    data <- MergeDims(data, merge_dims = spatial_dims, rename_dim = 'grid')
    spatial_dims <- 'grid'
  } else if (length(spatial_dims) > 2) {
    stop("Paramenter 'spatial_dims' maximum length 2.")
  }
  # time_dims
  if (!all(time_dims %in% names(dim(data)))) {
    stop("Parameter 'time_dims' is not found in 'data' dimension names. ", 
         "Please check the dimension names of 'data'.")
  }
  split_time <- NULL
  if (length(time_dims) == 2) {
    pos1 <- which(names(dim(data)) == time_dims[1])
    pos2 <- which(names(dim(data)) == time_dims[2])
    if (pos1 < pos2) {
      time_dims <- time_dims[c(1, 2)]
      split_time <- dim(data)[c(pos1, pos2)]
    } else {
      time_dims <- time_dims[c(2, 1)]
      split_time <- dim(data)[c(pos2, pos1)]
    }
    data <- MergeDims(data, merge_dims = time_dims, rename_dim = 'temporals')
    time_dims <- 'temporals'
  } else if (length(time_dims) > 2) {
    stop("Parameter 'time_dims' maximum length 2.")
  }
  if (any(is.na(data))) {
    warning("Parameter 'data' contains NAs.")
  }
  attractor <- Apply(data, target_dims = c(time_dims, spatial_dims),
                     fun = .proxiesattractor,
                     quanti = quanti , na.rm = na.rm, ncores = ncores)
  # rename dimensions
  attractor <- lapply(attractor, 
                      FUN = function(x, dimname) { 
                        names(dim(x))[dimname] <- time_dims
                        return(x)}, 
                      dimname = which(names(dim(attractor[[1]])) == 'dim2'))
  if (length(split_time) == 2) {
    attractor <- lapply(attractor, SplitDim, split_dim = time_dims,
                        indices = NULL, freq = split_time[1],
                        new_dim_name = names(split_time[2]))
    attractor <- lapply(attractor, function(x, name) {
                          names(dim(x))[which(names(dim(x)) == time_dims)] <-
                            names(split_time)[1]
                          return(x)})
  }
  # Reorder dimensions
  # attractor <- lapply(attractor, 
  #                     function(x) {
  #                       pos <- match(dimnames[!dimnames %in% input_spatial_dims], 
  #                                    names(dim(x)))
  #                       x <- aperm(x, pos)
  #                       return(x)
  #                     })
  return(list(dim = attractor$dim, theta = attractor$theta))
}

.proxiesattractor <- function(data, quanti, na.rm) {
  # expected dimensions data: time and grid
  grid_dim <- names(dim(data))[2]
  logdista <- Apply(data, target_dims =  grid_dim,
                    fun = function(x, y) {
                      -log(colMeans((y - as.vector(x))^2, na.rm = na.rm))},
                    y = t(data))[[1]]
  # Computation of theta
  names(dim(logdista)) <- c('dim1', 'dim2')
  proxies <- Apply(data = list(logdista = logdista),
                   target_dims = list('dim1'), fun = .Theta, quanti = quanti)
  
  return(list(dim = proxies$dim, theta = proxies$theta))
}

# Function to compute theta
.Theta <- function(logdista, quanti) {
  # Compute the thheshold corresponding to the quantile
  thresh  <- quantile(logdista, quanti, na.rm = TRUE)
  logdista[which(logdista == 'Inf')] <-  NaN
  Li <- which(as.vector(logdista) > as.numeric(thresh))
  #Length of each cluster
  Ti <- diff(Li)
  N <-  length(Ti)
  q <-  1 - quanti
  Si <- Ti - 1
  Nc <- length(which(Si > 0))
  N <- length(Ti)
  theta <- (sum(q * Si) + N + Nc - sqrt(((sum(q * Si) + N + Nc)^2) - 
            8 * Nc * sum(q * Si))) / (2 * sum(q * Si))
  #Sort the exceedances  
  logdista <- sort(logdista)
  #Find all the Peaks over Thresholds.
  findidx <- which(as.vector(logdista) > as.numeric(thresh))
  if (length(findidx) < 1) {
    stop("Parameter 'quanti' is too high for the length of the data provided.")
  }
  logextr <- logdista[findidx[[1]]:(length(logdista) - 1)]
  #The inverse of the dimension is just the average of the exceedances
  dim <- 1 /mean(as.numeric(logextr) - as.numeric(thresh))
  return(list(dim = dim, theta = theta))
}